from OpenGL.GL import *
from OpenGL.GL.shaders import compileProgram, compileShader

def create_shader_program(module_info: tuple[tuple[int, str]]) -> int:
    """
        Create an OpenGL shader program.

        Parameters:

            module_info: holds the source code and module types
    """

    modules = []

    for description in module_info:
        modules.append(create_shader_module(description))

    shader = compileProgram(*modules)

    for module in modules:
        glDeleteShader(module)

    return shader

def create_shader_module(module_info: tuple[int, str]) -> int:
    """
        Create an OpenGL shader module

        Parameters:

            module_info: describes the filenames and their module types.
    """

    module_type, src = module_info

    return compileShader(src, module_type)



import numpy as np

DATA_TYPE_TRANSFORM = np.dtype({
    'names': [
        'x', 'y', 'scale', 'rotation'],
    'formats': [
        np.float16, np.float16, 
        np.float16, np.float16],
    'offsets': [
        0, 2, 
        4, 6],
    'itemsize': 8})

DATA_TYPE_VERTEX = np.dtype({
    'names': [
        'imageSize_x','imageSize_y',
        'texoffset_x','texoffset_y', 'layer',
        'center_x', 'center_y', 
        'scale', 'rotation'],
    'formats': [
        np.float16, np.float16, 
        np.float16, np.float16, np.uint8, 
        np.float16, np.float16, 
        np.float16, np.float16],
    'offsets': [
        0, 2, 
        4, 6, 8, 
        9, 11, 
        13, 15],
    'itemsize': 17})


from OpenGL.GL import *
from PIL import Image

class Rect:
    __slots__ = (
        "width", "height", "top", 
        "left", "layer")

    def __init__(self, left: int, top: int, width: int, height: int, layer: int):

        self.width = width
        self.height = height
        self.top = top
        self.left = left
        self.layer = layer
    
    def split(self, x: int, y: int) -> tuple["Rect"]:

        x_over = self.width - x
        y_over = self.height - y

        self.width = x
        self.height = y

        return (
            Rect(self.left + x,     self.top, x_over,      y, self.layer), 
            Rect(    self.left, self.top + y,      x, y_over, self.layer), 
            Rect(self.left + x, self.top + y, x_over, y_over, self.layer))
    
    def get_area(self) -> int:

        return self.width * self.height
    
    def can_hold(self, w: int, h: int) -> bool:

        return (w <= self.width) and (h <= self.height)

    def __str__(self) -> str:

        return f"({self.left}, {self.top}, {self.width}, {self.height}, {self.layer})"

class TextureAtlas:
    """
        Manages a group of textures.
    """
    """
    __slots__ = (
        "_atlas", "_sprite_groups", "_dummy_vao", 
        "_shader", "_global_info_location",
        "_shader", "_dummy_vao", "_global_info_location")
        """

    
    def __init__(self):
        """
            Construct a new texture atlas.
        """

        self._max_w = 4096
        self._max_h = 4096
        self._history: dict[str, int] = {}
        self._free_rectangles: list[Rect] = []
        self._pending_allocations: list[Rect] = []
        self._allocated_rectangles: list[Rect] = []
        self.handle = 0
        self._layer_count = 0
        self._texture_pages: dict[int, Image.Image] = {}

    def load_image(self, filename: str) -> int:
        """
            Registers an image for loading.

            Parameters:

                filename: full filepath to the image to load.
            
            Returns:

                A handle to the loaded image, indicating the image's position
                within the set of loaded images.
        """

        if filename in self._history:
            return self._history[filename]

        i = len(self._history)
        self._history[filename] = i

        with Image.open(filename, mode = "r") as img:
            w, h = img.size
            self.allocate(Rect(0, 0, w, h, 0))

        return i
    
    def get_image_size(self, object_type: int) -> tuple[int]:
        rect = self._allocated_rectangles[object_type]
        return rect.width / 2, rect.height / 2
    
    def get_offset(self, object_type: int) -> tuple[int]:
        rect = self._allocated_rectangles[object_type]
        return rect.left, rect.top, rect.layer
    
    def get_max_size(self) -> tuple[int]:
        return self._max_w, self._max_h
    
    def build(self) -> None:

        if len(self._history) == 0:
            return
        
        # build opengl texture
        self.handle = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D_ARRAY, self.handle)
        glTexImage3D(GL_TEXTURE_2D_ARRAY, 0, GL_RGBA8, 
                    self._max_w, self._max_h, self._layer_count, 0, GL_RGBA, GL_UNSIGNED_BYTE, None)
        
        for filename, i in self._history.items():
            with Image.open(filename, mode = "r") as img:
                img = img.convert("RGBA")
                rect = self._allocated_rectangles[i]

                self._texture_pages[rect.layer].paste(img, (rect.left, rect.top))
        
        for layer, img in self._texture_pages.items():
            img_data = bytes(img.tobytes())
            w,h = img.size

            #img.save(f"page_{layer}.png")
                
            glTexSubImage3D(GL_TEXTURE_2D_ARRAY, 0, 
                            0, 0, layer, 
                            w, h, 1,
                            GL_RGBA,GL_UNSIGNED_BYTE,img_data)
            
            img.close()
        self._texture_pages = {}
        
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_R, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_T, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR)
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glGenerateMipmap(GL_TEXTURE_2D_ARRAY)
    
    def allocate(self, rect: Rect) -> None:

        if len(self._free_rectangles) == 0:
            self._free_rectangles.append(Rect(0, 0, self._max_w, self._max_h, self._layer_count))
            self._layer_count += 1
        
        #print(f"Rectangle: {str(rect)}")
        
        for candidate in self._free_rectangles:

            debug = candidate.layer == 0

            #if debug:
            #    print(f"Candidate: {str(candidate)}")

            if not candidate.can_hold(rect.width, rect.height):
                #if debug:
                #    print("cannot hold")
                continue
            
            #if debug:
            #    print("can hold")
            new_free_rects = candidate.split(rect.width, rect.height)
            self._free_rectangles.remove(candidate)
            self._allocated_rectangles.append(candidate)
            if candidate.layer not in self._texture_pages:
                self._texture_pages[candidate.layer] = Image.new(mode="RGBA", size=(self._max_w, self._max_h))

            for new_free_rect in new_free_rects:
                self.sorted_insert(new_free_rect, self._free_rectangles)
            
            return
        
        #print("new page")
        new_page = Rect(0, 0, self._max_w, self._max_h, self._layer_count)
        self._layer_count += 1
        new_free_rects = new_page.split(rect.width, rect.height)
        self._allocated_rectangles.append(new_page)

        if new_page.layer not in self._texture_pages:
            self._texture_pages[new_page.layer] = Image.new(mode="RGBA", size=(self._max_w, self._max_h))

        for new_free_rect in new_free_rects:
            self.sorted_insert(new_free_rect, self._free_rectangles)
    
    def sorted_insert(self, new_rect: Rect, rects: list[Rect]) -> None:

        area = new_rect.get_area()
        if area == 0:
            return

        rects.append(new_rect)
        """
        for i in range(len(rects) - 1, 1, -1):

            if rects[i - 1].get_area() > area:
                rects[i] = rects[i - 1]
            else:
                rects[i] = new_rect
                return
        
        rects[0] = new_rect
        """


import numpy as np
from .shaders import *
from .texture_atlas import TextureAtlas
from ..data_types import *
from krasue.config import *

class Renderer:
    """
        OpenGL 3.3 renderer. Can do instanced rendering but not indirect.
    """
    __slots__ = (
        "_atlas", "_sprite_groups", "_dummy_vao", 
        "_shader", "_global_info_location",
        "_shader", "_dummy_vao", "_global_info_location")
    

    def setup(self, width: int, height: int, title: str):
        """
            Builds a renderer

            Parameters:

                width, height: size of the window

                title: title for the window caption

            Returns:

                The window which will be rendered to
        """
        
        self._set_up_pygame(width, height, title)
        self._make_objects()
        self._set_up_opengl()

    def _set_up_pygame(self, width: int, height: int, title: str) -> any:
        """
            Initialize glfw, build a new window.

            Parameters:

                width, height: size of the window

                title: title for the window caption

            Returns:

                The window which will be rendered to
        """

        pg.init()
        pg.display.gl_set_attribute(pg.GL_CONTEXT_MAJOR_VERSION,3)
        pg.display.gl_set_attribute(pg.GL_CONTEXT_MINOR_VERSION,3)
        pg.display.gl_set_attribute(pg.GL_CONTEXT_PROFILE_MASK, 
                                    pg.GL_CONTEXT_PROFILE_CORE)
        pg.display.set_mode((width, height), pg.OPENGL|pg.DOUBLEBUF)
        pg.display.set_caption(title)
    
    def _make_objects(self) -> None:
        """
            Construct renderer objects
        """

        self._atlas = TextureAtlas()
        self._sprite_groups = []

        self._dummy_vao = 0
        self._shader = 0
        self._global_info_location = 0
    
    def _set_up_opengl(self) -> None:
        """
            Configure any one-time OpenGL setup.
        """

        glEnable(GL_STENCIL_TEST)
        glStencilMask(0xFF)
        glStencilFunc(GL_EQUAL, 0, 0xFF)
        glStencilOp(GL_KEEP, GL_INCR, GL_INCR)

    def set_clear_color(self, color: tuple[float]) -> None:
        """
            Sets the color with which to clear the screen upon update.

            Parameters:

                color: the desired clear color, in rgba form, where each
                    channel is a float in the range [0, 1.0]
        """

        glClearColor(*color)

    def load_image(self, filename: str) -> int:
        """
            Registers an image for loading.

            Parameters:

                filename: full filepath to the image to load.
            
            Returns:

                A handle to the loaded image, indicating the image's position
                within the set of loaded images.
        """

        return self._atlas.load_image(filename)
    
    def after_setup(self) -> None:
        """
            Upload all image handles to the GPU

            Parameters:

                window: the glfw window we'll be rendering to.
        """

        self._atlas.build()

        vertex_src = """
#version 330
uniform vec4 screenSize_maxSize;
layout(location=0) in vec2 imageSize;
layout(location=1) in vec2 texOffset;
layout(location=2) in float layer;
layout(location=3) in vec2 center;
layout(location=4) in float scale;
layout(location=5) in float rotation;

out vec3 fragTexCoord;

const vec2[4] coords = vec2[](
    vec2(-1, -1),
    vec2( 1, -1),
    vec2( 1,  1),
    vec2(-1,  1));

void main() {
    vec2 pos = coords[gl_VertexID];

    //scale
    pos.x = imageSize.x * scale * pos.x;
    pos.y = imageSize.y * scale * pos.y;

    //rotate
    float c = cos(radians(rotation));
    float s = sin(radians(rotation));
    float x = pos.x * c - pos.y * s;
    float y = pos.x * s + pos.y * c;
    pos.x = x;
    pos.y = y;

    //translate
    pos = pos + center;

    //convert to NDC
    pos.x = (pos.x - screenSize_maxSize.x) / screenSize_maxSize.x;
    pos.y = (pos.y - screenSize_maxSize.y) / screenSize_maxSize.y;

    gl_Position = vec4(pos, 0.0, 1.0);

    vec2 offset;
    offset.x = texOffset.x / 4096.0;
    offset.y = texOffset.y / 4096.0;
    vec2 size = 0.5 * (coords[gl_VertexID] + vec2(1.0));
    size.x = (size.x * imageSize.x) / screenSize_maxSize.z;
    size.y = (size.y * imageSize.y) / screenSize_maxSize.w;
    //pos.y = pos.y * -1;
    fragTexCoord = vec3(offset + size, layer);
}
"""

        fragment_src = """
#version 330
uniform sampler2DArray material;

in vec3 fragTexCoord;

out vec4 color;

void main() {
    vec4 sampled = texture(material, fragTexCoord);

    if (sampled.a < 0.1) {
        discard;
    }

    color = sampled;
}
"""
        info = (
            (GL_VERTEX_SHADER, vertex_src),
            (GL_FRAGMENT_SHADER, fragment_src)
        )
        self._dummy_vao = glGenVertexArrays(1)
        glBindVertexArray(self._dummy_vao)
        self._shader = create_shader_program(info)
        glUseProgram(self._shader)
        self._global_info_location = glGetUniformLocation(self._shader, "screenSize_maxSize")

        w,h = pg.display.get_window_size()
        max_w, max_h = self._atlas.get_max_size()
        global_info = np.array((w / 2, h / 2, max_w / 2, max_h / 2), dtype=np.uint32)
        glUniform4fv(self._global_info_location, 1, global_info)

    def start_drawing(self) -> None:
        """
            Perform any necessary setup before receiving draw commands
        """

        glClear(GL_COLOR_BUFFER_BIT | GL_STENCIL_BUFFER_BIT)

    def register_sprite_group(self, object_types: np.ndarray, 
                              transforms: np.ndarray, size: int) -> None:
        
        buffer = np.zeros(size, DATA_TYPE_VERTEX)
        for i in range(size):

            read_pos = size - 1 - i
            object_type = object_types[read_pos]
            
            w, h = self._atlas.get_image_size(object_type)
            buffer[i]['imageSize_x']   = w
            buffer[i]['imageSize_y']   = h

            x,y,z = self._atlas.get_offset(object_type)
            buffer[i]['texoffset_x']   = x
            buffer[i]['texoffset_y']   = y
            buffer[i]['layer']         = z

            buffer[i]['center_x']   = transforms[read_pos]['x']
            buffer[i]['center_y']   = transforms[read_pos]['y']
            buffer[i]['scale']      = transforms[read_pos]['scale']
            buffer[i]['rotation']   = transforms[read_pos]['rotation']
        
        VAO = glGenVertexArrays(1)
        glBindVertexArray(VAO)
        VBO = glGenBuffers(1)
        glBindBuffer(GL_ARRAY_BUFFER, VBO)

        stride = DATA_TYPE_VERTEX.itemsize
        offset = 0
        glBufferData(GL_ARRAY_BUFFER, size * stride, buffer, GL_STATIC_DRAW)

        # layout(location=0) in vec2 imageSize;
        glVertexAttribPointer(0, 2, GL_HALF_FLOAT, GL_FALSE, stride, ctypes.c_void_p(offset))
        glEnableVertexAttribArray(0)
        glVertexAttribDivisor(0,1)
        offset += 4

        # layout(location=1) in vec2 texOffset;
        glVertexAttribPointer(1, 2, GL_HALF_FLOAT, GL_FALSE, stride, ctypes.c_void_p(offset))
        glEnableVertexAttribArray(1)
        glVertexAttribDivisor(1,1)
        offset += 4

        # layout(location=2) in float layer;
        glVertexAttribPointer(2, 1, GL_UNSIGNED_BYTE, GL_FALSE, stride, ctypes.c_void_p(offset))
        glEnableVertexAttribArray(2)
        glVertexAttribDivisor(2,1)
        offset += 1

        # layout(location=3) in vec2 center;
        glVertexAttribPointer(3, 2, GL_HALF_FLOAT, GL_FALSE, stride, ctypes.c_void_p(offset))
        glEnableVertexAttribArray(3)
        glVertexAttribDivisor(3,1)
        offset += 4

        # layout(location=4) in float scale;
        glVertexAttribPointer(4, 1, GL_HALF_FLOAT, GL_FALSE, stride, ctypes.c_void_p(offset))
        glEnableVertexAttribArray(4)
        glVertexAttribDivisor(4,1)
        offset += 2

        # layout(location=5) in float rotation;
        glVertexAttribPointer(5, 1, GL_HALF_FLOAT, GL_FALSE, stride, ctypes.c_void_p(offset))
        glEnableVertexAttribArray(5)
        glVertexAttribDivisor(5,1)

        id = len(self._sprite_groups)
        self._sprite_groups.append((VAO, VBO, size))
        return id

    def draw_sprite_group(self, id: int) -> None:
        """
            Draw a set of sprites.

            Parameters:

                object_types: the image types for each sprite

                transform_infos: the transform for each sprite

                count: how many sprites to draw from the group
        """

        VAO, _ ,size = self._sprite_groups[id]
        glBindVertexArray(VAO)
        glDrawArraysInstanced(GL_TRIANGLE_FAN, 0, 4, size)

    def finish_drawing(self) -> None:
        """
            Called once per frame to draw stuff.
            Override this function to make your game draw things.
        """
        pg.display.flip()



import numpy as np
from .shaders import *
from .texture_atlas import TextureAtlas
from ..data_types import *
from krasue.config import *

class Renderer:
    """
        Around OpenGL 4.x, a lot of the developments were focussed on improving
        performance. These "Approaching Zero Driver Overhead" features
        allowed for faster drawing of arbitrary objects.
    """
    __slots__ = (
        "_image_history", "_image_sizes", "_image_gl_id", 
        "_max_image_w", "_max_image_h",
        "_shader", "_dummy_vao", "_global_info_location", 
        "_object_info_location", "_sprite_info_location")


    def setup(self, width: int, height: int, title: str):
        """
            Builds a renderer

            Parameters:

                width, height: size of the window

                title: title for the window caption

            Returns:

                The window which will be rendered to
        """
        
        glfw.init()
        glfw.window_hint(GLFW_CONSTANTS.GLFW_CONTEXT_VERSION_MAJOR,4)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_CONTEXT_VERSION_MINOR,5)
        glfw.window_hint(
            GLFW_CONSTANTS.GLFW_OPENGL_PROFILE, 
            GLFW_CONSTANTS.GLFW_OPENGL_CORE_PROFILE)
        glfw.window_hint(
            GLFW_CONSTANTS.GLFW_OPENGL_FORWARD_COMPAT, 
            GLFW_CONSTANTS.GLFW_TRUE)
        glfw.window_hint(
            GLFW_CONSTANTS.GLFW_DOUBLEBUFFER, 
            GLFW_CONSTANTS.GLFW_FALSE)
        glfw.window_hint(
            GLFW_CONSTANTS.GLFW_RESIZABLE,
            GLFW_CONSTANTS.GLFW_FALSE)
        
        window = glfw.create_window(width, height, title, None, None)
        glfw.make_context_current(window)

        #image loading stuff
        self._image_history: dict[str, int] = {}
        self._image_sizes = np.zeros(0, dtype = np.uint32)
        self._max_image_w = 0
        self._max_image_h = 0

        return window
    
    def set_clear_color(self, color: tuple[float]) -> None:
        """
            Sets the color with which to clear the screen upon update.

            Parameters:

                color: the desired clear color, in rgba form, where each
                    channel is a float in the range [0, 1.0]
        """

        glClearColor(*color)

    def load_image(self, filename: str) -> int:
        """
            Registers an image for loading.

            Parameters:

                filename: full filepath to the image to load.
            
            Returns:

                A handle to the loaded image, indicating the image's position
                within the set of loaded images.
        """

        if filename in self._image_history:
            return self._image_history[filename]

        i = len(self._image_history)
        self._image_history[filename] = i

        with Image.open(filename, mode = "r") as img:
            w, h = img.size
            self._image_sizes = np.append(self._image_sizes, (w,h))
            self._max_image_w = max(w, self._max_image_w)
            self._max_image_h = max(h, self._max_image_h)

        return i

    def after_setup(self, window) -> None:
        """
            Upload all image handles to the GPU
        """

        if len(self._image_history) > 0:

            self._image_gl_id = glGenTextures(1)
            glBindTexture(GL_TEXTURE_2D_ARRAY, self._image_gl_id)
            glTexStorage3D(GL_TEXTURE_2D_ARRAY, 1, GL_RGBA8, 
                        self._max_image_w, self._max_image_h, len(self._image_history))
            
            for filename, i in self._image_history.items():
                with Image.open(filename, mode = "r") as img:
                    w, h = img.size
                    data = bytes(img.convert("RGBA").tobytes())
                    glTexSubImage3D(GL_TEXTURE_2D_ARRAY, 0, 
                                    0, 0, i, 
                                    w, h, 0,
                                    GL_RGBA,GL_UNSIGNED_BYTE,data)
            
            glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_R, GL_REPEAT)
            glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_S, GL_REPEAT)
            glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_T, GL_REPEAT)
            glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
            glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
    
    def start_drawing(self) -> None:
        """
            Perform any necessary setup before receiving draw commands
        """

        glClear(GL_COLOR_BUFFER_BIT)

    def finish_drawing(self, window) -> None:
        """
            Called once per frame to draw stuff.
            Override this function to make your game draw things.
        """

        glFlush()


from .config import *
import numpy as np
import krasue.backends.opengl.modern as ogl_modern
import krasue.backends.opengl.azdo as ogl_azdo
from krasue.backends.data_types import *

import os
os.environ["SDL_VIDEO_X11_FORCE_EGL"] = "1"

BACKEND_AZDO_OGL = 0
BACKEND_MODERN_OGL = 1

RENDER_BEHAVIOR_EACH_FRAME = 0
RENDER_BEHAVIOR_CONSERVATIVE = 1

class Invocation:
    """
        Represents top level program control.
        More or less creates a window and rendering backend,
        and gives slots for behavior extension.
    """
    __slots__ = ("_renderer", )

    
    def __init__(self, width: int, height: int, 
                 title: str = "A spooky window",
                 backend: int = BACKEND_AZDO_OGL,
                 behavior: int = RENDER_BEHAVIOR_EACH_FRAME):
        """
            Invoke a krasue (very spooky, your exorcism license is
            up to date, right?)

            Parameters:

                width, height: size of the window

                title: title for the window caption

                backend: which sort of renderer the window
                    should use

                behavior: how hard to push the renderer.
                    Current options are to draw every frame
                    (good for understanding compute power, but
                    clogs the GPU with non-useful work)
                    or to render conservatively
                    (reduces non-visible renders, frees up CPU time
                    on main thread)
        """
        
        if (backend == BACKEND_AZDO_OGL and behavior == RENDER_BEHAVIOR_EACH_FRAME):
            self._renderer = ogl_azdo.Renderer()
            self._renderer.setup(width, height, title)
        if (backend == BACKEND_MODERN_OGL and behavior == RENDER_BEHAVIOR_EACH_FRAME):
            self._renderer = ogl_modern.Renderer()
            self._renderer.setup(width, height, title)
        
        self.on_setup()
        self._renderer.after_setup()

    def on_setup(self) -> None:
        """
            Override this method to implement any resource creation.
        """

        pass
    
    def new_object(self, filename: str) -> int:
        """
            Creates a new object.

            Parameters:

                filename: full filepath to associated image.
            
            Returns:

                The ID of the new object.
        """

        return self._renderer.load_image(filename)

    def set_clear_color(self, color: tuple[int]) -> None:
        """
            Sets the color with which to clear the screen upon update.

            Parameters:

                color: the desired clear color, in rgb form, where each
                    channel is an integer in the range [0, 255]
        """

        color = (
            min(1.0, max(0.0, color[0] / 255)), 
            min(1.0, max(0.0, color[1] / 255)), 
            min(1.0, max(0.0, color[2] / 255)), 1.0)
        self._renderer.set_clear_color(color)
    
    def set_title(self, title: str) -> None:
        """
            Set the title in the window's titlebar.

            Parameters:

                title: the title for the window.
        """

        pg.display.set_caption(title)

    def run(self) -> None:
        """
            Start the game's main loop.
        """

        running = True
        while (running):

            for event in pg.event.get():
                if event.type == pg.KEYDOWN \
                    and event.key == pg.K_ESCAPE:
                
                    running = False

            self.on_update()

            self._renderer.start_drawing()
            self.on_draw()
            self._renderer.finish_drawing()

    def on_update(self) -> None:
        """
            Called once per frame to update stuff (and things).
            Override this function to implement your game's update behavior.
        """

        pass

    def on_draw(self) -> None:
        """
            Called once per frame to draw stuff.
            Override this function to make your game draw things.
        """

        pass

class SpriteGroup:
    """
        A group of sprites, can hold an arbitrary number of sprites
        of different types.
    """
    __slots__ = (
        "_renderer", "_object_types", "_transforms",
        "_size", "_capacity", "_id")

    
    def __init__(self, invocation: Invocation):
        """
            Invoke a krasue (very spooky, your exorcism license is
            up to date, right?)

            Parameters:

                width, height: size of the window

                title: title for the window caption

                backend: which sort of renderer the window
                    should use

                behavior: how hard to push the renderer.
                    Current options are to draw every frame
                    (good for understanding compute power, but
                    clogs the GPU with non-useful work)
                    or to render conservatively
                    (reduces non-visible renders, frees up CPU time
                    on main thread)
        """
        
        self._renderer = invocation._renderer
        
        self._object_types = np.zeros(1, dtype=np.uint32)
        self._transforms = np.zeros(1, dtype=DATA_TYPE_TRANSFORM)

        self._size = 0
        self._capacity = 1
    
    def add(self, object_type: int, 
            x: float = 0.0, y: float = 0.0, 
            scale: float = 1.0, rotate: float = 0.0) -> int:
        """
            Adds a new sprite to this group

            Parameters:

                object_type: the image index representing the sprite
                x,y: center position of the sprite, default is (0,0)
                scale: scale of the sprite, default is 1.0
                rotate: angle of the sprite, default is 0.0
            
            Returns:

                The index of the sprite within the group, this can be used to
                later remove the sprite.
        """

        #position to insert at
        i = self._size

        #resize if needed
        if self._size >= self._capacity:
            new_object_types = np.zeros(self._capacity * 2, dtype=np.uint32)
            new_object_types[:self._size] = self._object_types[:]
            self._object_types = new_object_types

            new_transforms = np.zeros(self._capacity * 2, dtype=DATA_TYPE_TRANSFORM)
            new_transforms[:self._size] = self._transforms[:]
            self._transforms = new_transforms

            self._capacity = len(self._object_types)
        
        #insert
        self._object_types[i] = object_type
        self._transforms[i]['x'] = x
        self._transforms[i]['y'] = y
        self._transforms[i]['scale'] = scale
        self._transforms[i]['rotation'] = rotate
        self._size += 1

        return i

    def remove(self, i: int):
        """
            destroys the sprite at index i, maintains the order of the
            other sprites in the group.
        """
        
        np.delete(self._object_types, i)
        np.delete(self._transforms, np.s_[4 * i : 4 * (i + 1)])
        self._size -= 1

    def draw(self) -> None:
        """
            draw the sprite group
        """

        self._renderer.draw_sprite_group(self._id)

    def inscribe(self) -> None:
        """
            Register this sprite group with the renderer, duplicates info and uploads
            it to the GPU.
        """

        self._id = self._renderer.register_sprite_group(
            self._object_types, self._transforms, self._size)